#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <pcre.h>
#include <gtk/gtk.h>

#include "packages.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

static void pkg_entry_class_init (PkgEntryClass *klass);
static void pkg_entry_init (PkgEntry *entry);
static void pkg_entry_finalize (GObject *object);

static void pkg_entry_set_property (GObject            *obj,
                                     guint              prop_id,
                                     const GValue       *val,
                                     GParamSpec         *pspec);
static void pkg_entry_get_property (GObject    *obj,
                                     guint      prop_id,
                                     GValue     *val,
                                     GParamSpec *pspec);

static void pkg_entry_set_uri (PkgEntry *entry,
                                const gchar *uri);
static void pkg_entry_add_file (PkgEntry *entry,
                                 const gchar *file);
static void pkg_entry_set_name (PkgEntry *entry,
                                 const gchar *name) ;
static void pkg_entry_set_version (PkgEntry *entry,
                                    const gchar *version);
static void pkg_entry_set_arch (PkgEntry *entry,
                                 const gchar *arch);
static void pkg_entry_set_release (PkgEntry *entry,
                                    const gchar *release);
static void pkg_entry_set_description (PkgEntry *entry,
                                        const gchar *description);

enum {
  PROP_URI = 1,
  PROP_NAME,
  PROP_VERSION,
  PROP_ARCH,
  PROP_RELEASE,
  PROP_PKGNAME,
  PROP_SIZE,
  PROP_DESCRIPTION,
  PROP_SUMMARY,
  PROP_MTIME,
};

static GObjectClass *parent_class = NULL;

GType
pkg_entry_get_type (void)
{
  static GType type = 0;

  if (type == 0) {
    static const GTypeInfo info = {
      sizeof (PkgEntryClass),
      NULL,   /* base_init */
      NULL,   /* base_finalize */
      (GClassInitFunc) pkg_entry_class_init,   /* class_init */
      NULL,   /* class_finalize */
      NULL,   /* class_data */
      sizeof (PkgEntry),
      0,      /* n_preallocs */
      (GInstanceInitFunc) pkg_entry_init    /* instance_init */
    };

    type = g_type_register_static (G_TYPE_OBJECT, "PkgEntry", &info, 0);
  }
  return type;
}

static void
pkg_entry_class_init (PkgEntryClass *klass)
{
  GObjectClass *object_class;

  parent_class = (GObjectClass*) g_type_class_peek_parent (klass);
  object_class = (GObjectClass*) klass;

  object_class->finalize = pkg_entry_finalize;
	
  object_class->set_property = pkg_entry_set_property;
  object_class->get_property = pkg_entry_get_property;	
	
  g_object_class_install_property (object_class, PROP_URI,
                                   g_param_spec_string ("uri",
                                                        "URI",
                                                        "Package logfile URI",
                                                        "",
                                   G_PARAM_READWRITE|G_PARAM_CONSTRUCT_ONLY));	
  g_object_class_install_property (object_class, PROP_NAME,
                                   g_param_spec_string ("name",
                                                        "Name",
                                                        "Package name",
                                                        "",
                                                        G_PARAM_READWRITE));	
  g_object_class_install_property (object_class, PROP_VERSION,
                                   g_param_spec_string ("version",
                                   "Version",
                                   "Package version",
                                   "",
                                   G_PARAM_READWRITE));	
  g_object_class_install_property (object_class, PROP_ARCH,
                                   g_param_spec_string ("arch",
                                   "Architecture",
                                   "Package target architecture",
                                   "",
                                   G_PARAM_READWRITE));	
  g_object_class_install_property (object_class, PROP_RELEASE,
                                   g_param_spec_string ("release",
                                                        "Release",
                                                        "Package release id",
                                                        "",
                                                        G_PARAM_READWRITE));	
  g_object_class_install_property (object_class, PROP_PKGNAME,
                                   g_param_spec_string ("package-name",
                                                        "Package name",
                                                        "Full package name",
                                                        "",
                                                        G_PARAM_READABLE));	
  g_object_class_install_property (object_class, PROP_DESCRIPTION,
                                   g_param_spec_string ("description",
                                                        "Package description",
                                                        "Package description",
                                                        "",
                                                        G_PARAM_READWRITE));	
  g_object_class_install_property (object_class, PROP_SUMMARY,
                                   g_param_spec_string ("summary",
                                                        "Package summary",
                                                        "Package summary",
                                                        "",
							G_PARAM_READABLE));	
  g_object_class_install_property (object_class, PROP_SIZE,
                                   g_param_spec_long ("size",
                                                      "Package size",
                                                      "Package size on disk",
                                                      0, G_MAXLONG, 0,
                                                      G_PARAM_READWRITE));	
  g_object_class_install_property (object_class,
                                   PROP_MTIME,
                                   g_param_spec_long ("mtime",
                                                      "Package modification time",
                                                      "Package modification time",
                                                      1, G_MAXLONG, 1,
                                                      G_PARAM_READWRITE));
}

static void
pkg_entry_init (PkgEntry *entry)
{
  g_assert (entry != NULL);
  g_assert (IS_PKG_ENTRY (entry));
	
  entry->uri = g_strdup("");
  entry->name = g_strdup ("");
  entry->version = g_strdup ("");
  entry->arch = g_strdup ("");
  entry->release = g_strdup ("");
  
  entry->size = 0;
  entry->summary = g_strdup("");
  entry->description = g_strdup("");

  entry->files = NULL;
  entry->pixbuf = NULL;
  
  entry->mtime = 0;
}

static void
pkg_entry_finalize (GObject *object)
{
  PkgEntry *entry = PKG_ENTRY (object);
  GList *l;
  
  /* free all records and free all memory used by the entry */
  g_free (entry->uri);
  g_free (entry->name);
  g_free (entry->version);
  g_free (entry->release);
  for(l=entry->files; l; l=l->next)
      g_free(l->data);

  g_list_free(entry->files);	
  (* parent_class->finalize) (object);
}

PkgEntry*
pkg_entry_new (const gchar *uri)
{
  return PKG_ENTRY (g_object_new (PKG_ENTRY_TYPE, "uri", uri, NULL));
}

static void
pkg_entry_set_property ( GObject        *obj,
                         guint          prop_id,
                         const GValue   *val,
                         GParamSpec     *pspec)
{
  PkgEntry *entry;
  entry = PKG_ENTRY (obj);
  
  switch (prop_id)
    {
      case PROP_URI:
          pkg_entry_set_uri (entry, g_value_get_string (val));
          break;
      case PROP_NAME:
          pkg_entry_set_name (entry, g_value_get_string (val));
          break;
      case PROP_VERSION:
          pkg_entry_set_version (entry, g_value_get_string (val));
          break;
      case PROP_ARCH:
          pkg_entry_set_arch (entry, g_value_get_string (val));
          break;
      case PROP_RELEASE:
          pkg_entry_set_release (entry, g_value_get_string (val));
          break;
      case PROP_PKGNAME:
          g_warning ("property \"package-name\" is read-only");
          break;
      case PROP_SUMMARY:
          g_warning ("property \"summary\" is read-only");
          break;
      case PROP_SIZE:
          entry->size = g_value_get_long (val);
          break;
      case PROP_MTIME:
          entry->mtime = g_value_get_long (val);
          break;
      default:
          G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
          break;
    }
}

static void
pkg_entry_get_property (GObject         *obj,
                        guint           prop_id,
                        GValue          *val,
                        GParamSpec      *pspec)
{
  PkgEntry *entry;
  entry = PKG_ENTRY (obj);
  gchar *str;
  gchar *p;

  switch (prop_id)
    {
    case PROP_URI:
      g_value_set_string (val, entry->uri);
      break;
    case PROP_NAME:
      g_value_set_string (val, entry->name);
      break;
    case PROP_VERSION:
      g_value_set_string (val, entry->version);
      break;
    case PROP_ARCH:
      g_value_set_string (val, entry->arch);
      break;
    case PROP_RELEASE:
      g_value_set_string (val, entry->release);
      break;
    case PROP_PKGNAME:
      str = g_strdup_printf ("%s-%s-%s-%s", entry->name, entry->version, entry->arch, entry->release);
      g_value_set_string (val, str);
      g_free (str);
      break;
    case PROP_DESCRIPTION:
      g_value_set_string (val, entry->description);
      break;
    case PROP_SUMMARY:
      str = g_strdup (entry->description);
      p = index (str, '\n');
      if (p != NULL)
          *p = '\0';
              
      if (strlen (str) > 0)
          g_value_set_string (val, str);
      else
          g_value_set_string (val, "(no description available)");
                      
      g_free (str);
      break;
    case PROP_SIZE:
      g_value_set_long (val, entry->size);
      break;
    case PROP_MTIME:
      g_value_set_long (val, entry->mtime);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
      break;
    }
}

static gboolean get_pixbuf_from_desktop (PkgEntry *entry,
                                         const gchar *filename);
static gint parse_section_start (PkgEntry *entry,
                                 const gchar *tag,
                                 const gchar *value);

static void
pkg_entry_add_file (PkgEntry *entry, const gchar *file)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (file != NULL);
	
  gchar *tmp;
  gchar *p;
  tmp = g_strdup (file);
  p = index (tmp, '\n');
  if (p)
      *p = '\0';
  entry->files = g_list_append (entry->files, tmp);
        
  gchar *z, *pd, *str;
  str = g_strdup (file);
        
  if ((z = rindex (str, '/')) != NULL)
    {
      z++;
      pd = g_strdup_printf ("%s.desktop\n", entry->name);
      if (strcmp (z, pd) == 0)
        {
          if (get_pixbuf_from_desktop (entry, str))
              entry->score = 8;
        }
      else if ((fnmatch ("*.desktop", str, 0) == 0) && (entry->score < 4))
        {
          if (get_pixbuf_from_desktop (entry, str))
              entry->score = 4;
        }
        
      g_free (pd);
    }
}

static void
pkg_entry_set_name (PkgEntry *entry, const gchar *name)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (name != NULL);
	
  g_free (entry->name);
  entry->name = g_strdup (name);
}

static void
pkg_entry_set_version (PkgEntry *entry, const gchar *version)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (version != NULL);
	
  g_free (entry->version);
  entry->version = g_strdup (version);
}

static void
pkg_entry_set_arch (PkgEntry *entry, const gchar *arch)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (arch != NULL);
	
  g_free (entry->arch);
  entry->arch = g_strdup (arch);
}

static void
pkg_entry_set_release (PkgEntry *entry, const gchar *release)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (release != NULL);
	
  g_free (entry->release);
  entry->release = g_strdup (release);
}

static void
pkg_entry_set_description (PkgEntry *entry, const gchar *description)
{
	g_return_if_fail (IS_PKG_ENTRY (entry));
	g_return_if_fail (description != NULL);
	
	GString *str = g_string_new ("");
	gchar *descr = g_strdup (description);
	gchar *p = descr;
	gchar c[3] = { '\0', '\0' };
	
	while (*p) {
		switch (*p) {
			case '&':
				str = g_string_append (str, "&amp;");
				break;
			case '<':
				str = g_string_append (str, "&lt;");
				break;
			case '>':
				str = g_string_append (str, "&gt;");
				break;
			default:
				*c = *p;
				str = g_string_append (str, c);
			 break;
		}
		p++;
	}
	entry->description = g_strdup (str->str);
	g_string_free (str, TRUE);
	g_free (descr);
}

enum {
  PKG_SECTION_NONE = 0,
  PKG_SECTION_NAME,
  PKG_SECTION_USIZE,
  PKG_SECTION_CSIZE,
  PKG_SECTION_DESCR,
  PKG_SECTION_FILES,
  PKG_SECTION_OTHER,
};

static gint
get_pkg_section_id (const gchar *str)
{
  g_return_val_if_fail (str != NULL, PKG_SECTION_NONE);
  g_return_val_if_fail (*str != '\0', PKG_SECTION_NONE);
	
  if (strcmp (str, "PACKAGE NAME") == 0)
      return PKG_SECTION_NAME;
  else if (strcmp (str, "UNCOMPRESSED PACKAGE SIZE") == 0)
      return PKG_SECTION_USIZE;
  else if (strcmp (str, "PACKAGE DESCRIPTION") == 0)
      return PKG_SECTION_DESCR;
  else if (strcmp (str, "FILE LIST") == 0)
      return PKG_SECTION_FILES;
  else
      return PKG_SECTION_OTHER;
}	

/* this is, basically, package initialization since URI can only
   be set at PkgEntry creation */
static void
pkg_entry_set_uri (PkgEntry *entry, const gchar *uri)
{
  g_return_if_fail (IS_PKG_ENTRY (entry));
  g_return_if_fail (uri != NULL);
  
  printf ("Reading %s\n", uri);
	
  FILE *fd; /* FIXME: souldn't we use gnomevfs instead? */
	
  pcre *re;
  int rc;
  char *error;
  int erroffset;
  int ovector[16];
	
  gchar *str;
  GString *descr;
	
  /* FIXME: souldn't we use gnomevfs instead? */
  if ((fd = fopen(uri, "r")) != NULL) 
    { 
      str = g_malloc (256);
      gchar *re_str;
      gint current_section = PKG_SECTION_NONE;
      descr = g_string_new ("");
      gchar *tmp = NULL;
		
      while (fgets (str, 255, fd))
        {
          re_str = g_strdup ("^\\s*(?P<tag>[A-Z\\-\\ ]+)\\s*\\:\\s*(?P<val>.*)\\s*$");
          
          if ((re = pcre_compile (re_str, 0, (const char **) &error,
                                  &erroffset, NULL)) == NULL)
            {
              g_free (re_str);
              goto pcre_error;
            }
            
          if ((rc = pcre_exec (re, NULL, str, strlen (str),
                               0, 0, ovector, 16)) < 1)
            {
              /* this is a multi-line value continued */
              switch (current_section)
                {
                  case PKG_SECTION_DESCR:
                    if (tmp == NULL)
                        tmp = g_strdup_printf ("%s:", entry->name);
                    
                    if (strncmp (str, tmp, strlen (tmp)) == 0)
                        descr = g_string_append (descr, &str[strlen (tmp)]);
                    else
                        descr = g_string_append (descr, str);
                    break;
                  case PKG_SECTION_FILES:
                    pkg_entry_add_file (entry, str);
                    break;
                  default:
                    g_warning ("%s: skipping \"%s\"", uri, str);
                    break;
                }
            }
          else
            {
              gchar *ptag = g_malloc (strlen (str));
              gchar *pval = g_malloc (strlen (str));
              gint ntag, nval;
                                                
              ntag = pcre_get_substring (str, ovector, rc,
                                         pcre_get_stringnumber (re, "tag"),
                                         (const char **) &ptag);
              nval = pcre_get_substring (str, ovector, rc,
                                         pcre_get_stringnumber (re, "val"),
                                         (const char **) &pval);
              
              if (ntag <1)
                {
                  g_free (ptag);
                  g_free (pval);
                  goto parse_error;
                }

              if ( (current_section = parse_section_start (entry, ptag, pval)) /* FIXME? */
                   < 0 )
                {
                  g_free (pval);
                  g_free (ptag);
                  
                  goto parse_error;
                }

              /* FIXME: to be pedantic, we should treat possible value here as
                        the 1st line of a multi-line value */
            }
            
          g_free (re_str);
          pcre_free (re);
        } /* while (fgets ()) */
        
      pkg_entry_set_description (entry, descr->str);
		
      if (entry->pixbuf == NULL)
        {
          if (fnmatch ("gnome*", entry->name, 0) == 0)
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/gnome-package.png", NULL);
          else if (fnmatch ("gtk*", entry->name, 0) == 0)
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/gtk-package.png", NULL);
          else if (fnmatch ("kde*", entry->name, 0) == 0)
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/kde-package.png", NULL);
          else if (fnmatch ("kernel*", entry->name, 0) == 0)
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/kernel-package.png", NULL);
          else if (fnmatch ("*lib*", entry->name, 0) == 0)
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/lib-package.png", NULL);
          else
              entry->pixbuf = gdk_pixbuf_new_from_file (
                              PACKAGE_PIXMAPS_DIR"/default-package.png", NULL);
        }

      g_free (str);
      fclose (fd);
        
      g_free (entry->uri);
      entry->uri = g_strdup (uri);
      
      /* read file mtime */
      struct stat *ss;
      ss = g_new0 (struct stat, 1);
        
      if (stat (uri, ss) == -1) /* couldn't stat */
        {
          perror ("stat ()");
          entry->mtime = 0;
        }
      else
          entry->mtime = ss->st_mtime;

      g_free (ss);
    } /* fopen() */
     
  else
      g_warning ("Could not read %s", uri);

  return;

pcre_error:
  g_free (re);
  g_error ("Could not parse package logfile %s: pcre error", uri);
  return;

parse_error:
  g_free (str);
  g_error ("Could not parse package logfile %s: invalid package file", uri);
  return;
}

static gint
parse_section_start (PkgEntry *entry, const gchar *tag, const gchar *value)
{
  g_assert (IS_PKG_ENTRY (entry));
  g_return_val_if_fail (tag != NULL, -1);
  
  gint section = get_pkg_section_id (tag);
  
  pcre *re;
  int ovector[16];
  char *error;
  int erroffset;
  int rc;
  gchar *regex_str;
                                                    
  if (section == PKG_SECTION_NAME)
    {
      g_return_val_if_fail (value != NULL, -1);
      g_return_val_if_fail (strlen (value), -1);
        
      regex_str = g_strdup ("(?P<name>.+)\\-(?P<ver>[^\\-]+)\\-(?P<arch>[^\\-]+)\\-(?P<rel>[^\\-\\s]+)");
      
      if (!(re = pcre_compile (regex_str, 0, (const char **) &error,
                               &erroffset, NULL)))
        {
          g_free (regex_str);
          return -1;
        }
    
      if ((rc = pcre_exec (re, NULL, value, strlen (value),
                           0, 0, ovector, 16)) < 1)
        {
          pcre_free (re);
          g_free (regex_str);
          return -1;
        }
                              
      g_free (regex_str);
      gchar *pname = g_malloc (strlen (value));
      gchar *pver = g_malloc (strlen (value));
      gchar *prel = g_malloc (strlen (value));
      gchar *parch = g_malloc (strlen (value));
      int nname, nver, nrel, narch;
                              
      nname =   pcre_get_substring (value, ovector, rc,
                                    pcre_get_stringnumber (re, "name"),
                                    (const char **) &pname);
      nver =    pcre_get_substring (value, ovector, rc,
                                    pcre_get_stringnumber (re, "ver"),
                                    (const char **) &pver);
      nrel =    pcre_get_substring (value, ovector, rc,
                                    pcre_get_stringnumber (re, "rel"),
                                    (const char **) &prel);
      narch =   pcre_get_substring (value, ovector,
                                    rc, pcre_get_stringnumber (re, "arch"),
                                    (const char **) &parch);
                              
      if ((nname < 0) || (nver < 0) || (nrel < 0))
        {
          g_free (pname);
          g_free (pver);
          g_free (prel);
          g_free (parch);
          pcre_free (re);
          return -1;
        }
          
      pkg_entry_set_version (entry, pver);
      pkg_entry_set_release (entry, prel);
      pkg_entry_set_name (entry, pname);
      pkg_entry_set_arch (entry, parch);

      g_free (pname);
      g_free (pver);
      g_free (prel);
      g_free (parch);
      pcre_free (re);
    } /* PKG_SECTION_NAME: */
    
  else if (section == PKG_SECTION_USIZE)
    {
      g_return_val_if_fail (value != NULL, -1);
      g_return_val_if_fail (strlen (value), -1);
          
      glong size = 0;
      gchar *mod = g_malloc (strlen (value));
                  
      if (sscanf (value, "%ld %c", &size, mod) == 2)
        {
          if ((*mod == 'k') || (*mod == 'K'))
              size = size * 1024;
          else if ((*mod == 'm') || (*mod == 'M'))
              size = size * 1048576;
                      
          entry->size = size;
        }
                    
        g_free (mod);
    } /* PKG_SELECTION_USIZE */
    
  return section;
}

static gboolean
get_pixbuf_from_desktop (PkgEntry *entry, const gchar *filename)
{
  gchar *f, *p, *str;
  FILE *desktop;
  gchar *fn = g_strdup (filename);

  p = index (fn, '\n');
  if (p)
    *p = '\0';
	
  if ( fn[0] != '/' )
    f = g_strdup_printf ("/%s", fn);
  else
    f = g_strdup (fn);
	
  g_free (fn);
	
  if ( (desktop = fopen (f, "r")) == NULL )
    {
      g_free (f);
      return FALSE;
    }
  else
      g_free(f);
	
  str = g_malloc (256);
  gboolean go = TRUE;
  while (((str = fgets (str, 255, desktop)) != NULL) && go)
    {
      if (fnmatch ("Icon=*", str, 0) == 0)
        {
          f = &str[5];
          if (index (f, '/'))
            {
              if ((p = index (f, '\n')))
                  *p = '\0';
              if (entry->pixbuf != NULL)
                  g_free (entry->pixbuf);
                  entry->pixbuf = gdk_pixbuf_new_from_file (f, NULL);

              if(entry->pixbuf)
                {
                  go = FALSE;
                }
            }
          else
            {
              if ((p = index (f, '\n')))
                  *p = '\0';
              p = index (f, '.');

              if(p && (strcmp (".png", p) == 0))
                  *p = '\0';
              else if(p && (strcmp (".gif", p) == 0))
                  *p = '\0';
              else if(p && (strcmp (".xpm", p) == 0))
                  *p = '\0';
              if (entry->pixbuf != NULL)
                  g_free (entry->pixbuf);
              
              entry->pixbuf =
                  gtk_icon_theme_load_icon (gtk_icon_theme_get_default(),
                                            f, 48, 0, NULL);
              go = FALSE;
            }
        }
    }
    
  g_free (str);
  fclose (desktop);
    
  if(entry->pixbuf != NULL) 
      return TRUE;
  else
      return FALSE;
}
